#!/usr/bin/env python

"""Generate One Time Passwords ala rfc2289

$Id: otp.py,v 1.8 2005/10/11 18:58:02 sjg Exp $
"""

import getopt, os, sys
#import array
import struct
import md5, sha
import getpass
from utils import *

class OTP:
    def __init__(self, conf={}):
        self.conf = conf
        self.debug = getConfig(conf, 'debug', 0)
        self.opie = getConfig(conf, 'opie', False)
        self.pw = getConfig(conf, 'pw', None)
        self.key = None
        # we do this rather than subclass so that all our
        # subclasses can support both
        alg = getConfig(conf, 'hash', 'sha')
        if alg.lower() == 'md5':
            self.alg = 'MD5'
            self.f_hash = self.f_md5
        else:
            self.alg = 'SHA1'
            self.f_hash = self.f_sha
            
    def getPassphrase(self):
        pw = getpass.getpass("Using %s - Enter pass-phrase: " % self.alg)
        if sys.stdin.isatty():
            print "\n"
        return pw

    def f_md5(self, data):
        """Hash data using MD5 and fold to 64bits"""
        hash = md5.new()

        hash.update(data)
        if self.debug > 3:
            print "data='%s' ->" % data, hash.hexdigest()
        fmt = '<LLLL'
        w = struct.unpack(fmt, hash.digest())

        if self.debug > 1:
            print "w=", map(hex, w)
        # fold per rfc2289 (and rfc1938)
        d = []
        d.append(w[0] ^ w[2])
        d.append(w[1] ^ w[3])
        return d

    def f_sha(self, data):
        """Hash data using SHA1 and fold to 64bits

        SHA1 is supposed to be a big-endian algorithm,
        yet OPIE seems to treat it as little endian, or
        perhaps that is an artifact of the SHA1 implementation on FreeBSD.

        Regardless, if the opie flag is set we unpack the digest as
        little endian - which gives us the same result as OPIE (rfc1938),
        otherwise we unpack it as big endian giving us the same
        results as expected by rfc2289.
        """
        hash = sha.new()

        hash.update(data)
        if self.debug > 3:
            print "data='%s' ->" % data, hash.hexdigest()
        if self.opie:
            fmt = '<LLLLL'              # OPIE uses rfc 1938
        else:
            fmt = '>LLLLL'              # rfc 2289
            
        w = struct.unpack(fmt, hash.digest())

        if self.debug > 1:
            print "w=", map(hex, w)
        # fold per rfc2289 (and rfc1938)
        d = []
        d.append(w[0] ^ w[2] ^ w[4])
        d.append(w[1] ^ w[3])
        return d

    def f(self, data):
        """Hash input using the selected algorithm"""
        d = self.f_hash(data)
        if self.debug > 1:
            print "d=", map(hex, d)
        # this may not work on big-endian machine...
        # return array.array('L', d).tostring()
        res = ''
        i = 0;
        while i < 2:
            res += '%c' % (d[i] & 0xff)
            res += '%c' % ((d[i] >> 8) & 0xff)
            res += '%c' % ((d[i] >> 16) & 0xff)
            res += '%c' % ((d[i] >> 24) & 0xff)
            i += 1
        return res

    def renderPassword(self, x):
        return x
        
    def genPassword(self, n, seed):
        if not self.key:
            if self.pw:
                pw = self.pw
                self.pw = ''
            else:
                pw = self.getPassphrase()
            self.key = self.f(seed + pw)
        x = self.key
        while n > 0:
            x = self.f(x)
            n -= 1;
        w = struct.unpack('BBBBBBBB', x)
        x = b2x(w, 2)
        if self.debug:
            print "w=", map(hex, w)
            print "x=", x
        return self.renderPassword(x)

    
if __name__ == '__main__':
    """One Time Password using SHA1 and MD5"""
    debug = 0
    doHex = 0
    conf = {}
    alg = 'sha'
    count = 1
    myname = os.path.basename(sys.argv[0])
    i = myname.rfind('-')
    if i > 0:
        alg = myname[i+1:]

    try:
        opts, args = getopt.getopt(sys.argv[1:], "Dt:Oxn:p:")
    except getopt.GetoptError:
        print >> sys.stderr, "Use the source Luke!"
        sys.exit(2)
    for o, a in opts:
        if o == '-D':
            conf['pw'] = "OTP's are good"
            debug += 1
        elif o == '-t':
            alg = a
        elif o == '-O':
            conf['opie'] = True
        elif o == '-x':
            doHex = 1
        elif o == '-n':
            count = int(a)
        elif o == '-p':
            conf['pw'] = a

    conf['debug'] = debug
    conf['hash'] = alg
    if doHex:
        otp = OTP(conf)
    else:
        import otpw
        otp = otpw.OTPWords(conf)

    if debug:
        x = otp.genPassword(99, 'correct')
        print x
        sys.exit(0)

    if len(sys.argv) < 3:
        print "Usage:\n\t%s count seed\n" % sys.argv[0]
        sys.exit(1)

    seq = int(args[0])
    seed = args[1]
    if count == 1:
        x = otp.genPassword(seq, seed)
        print x
    else:
        seq -= count
        while count > 0:
            seq += 1
            x = otp.genPassword(seq, seed)
            print '%4d: %s' % (seq, x)
            count -= 1

